from graphics_and_games_klassen import Kreis, Text, Rechteck, Dreieck, Zeichenfenster, Figur
import math, sqlite3, random
    

# Verwaltet ein Knotensymbol in einem Graph
class KnotenSymbol():

    # Legt das Symbol an und besetzt die Attribute.
    def __init__(self, x, y, radius, farbe, bezeichner):
        self.x = x
        self.y = y
        self.r = radius
        self.farbe = farbe
        self.bezeichner = bezeichner
        self.aussen = Kreis()
        self.aussen.FarbeSetzen("schwarz")
        self.innen = Kreis()
        self.text = Text()
        self.text.TextGroesseSetzen(18)
        self.DarstellungAktualisieren()

    # Aktualisiert das Symbol
    def DarstellungAktualisieren(self):
        self.aussen.PositionSetzen(self.x, self.y)
        self.aussen.RadiusSetzen(self.r)
        self.innen.PositionSetzen(self.x, self.y)
        self.innen.RadiusSetzen(self.r - 2)
        self.innen.FarbeSetzen(self.farbe)
        self.text.PositionSetzen(int(self.x - len(self.bezeichner) / 2 * 10 - len(self.bezeichner) % 2 * 5), self.y - 6)
        self.text.TextSetzen(self.bezeichner)

    # Setzt die Farbe der Darstellung
    def FarbeSetzen(self, f):
        self.farbe = f
        self.DarstellungAktualisieren()

    # Setzt die Position der Darstellung
    def PositionSetzen(self, x, y):
        self.x = x
        self.y = y
        self.DarstellungAktualisieren()

    # Setzt die Farbe der Darstellung
    def BezeichnerSetzen(self, bezeichner):
        self.bezeichner = bezeichner
        self.DarstellungAktualisieren()

    # Meldet die x-Koordinate des Symbols zurück.
    def XGeben(self):
        return self.x

    # Meldet die y-Koordinate des Symbols zurück.
    def YGeben(self):
        return self.y

    # Meldet den Bezeichner des Symbols zurück.
    def BezeichnerGeben(self):
        return self.bezeichner

    # Entfernt das Knotensymbol aus der Anzeige
    def Entfernen(self):
        self.aussen.Entfernen()
        self.innen.Entfernen()
        self.text.Entfernen()
    
    # Setzt die Sichtbarkeit des Symbols
    def SichtbarkeitSetzen(self, sichtbar):
        self.aussen.SichtbarkeitSetzen(sichtbar)
        self.innen.SichtbarkeitSetzen(sichtbar)
        self.text.SichtbarkeitSetzen(sichtbar)

# Verwaltet ein Kantensymbol in einem Graph
class KantenSymbol():

    # Legt das Symbol an und besetzt die Attribute.
    def __init__(self, start, ende, gerichtet, gewicht, breite, farbe):
        self.von = start
        self.nach = ende
        self.farbe = farbe
        self.breite = breite
        self.gerichtet = gerichtet
        self.gewicht = gewicht
        self.aussen = Rechteck()
        self.innen = Rechteck()
        self.pfeil = Dreieck()
        self.pfeil.FarbeSetzen("schwarz")
        self.text = Text()
        self.text.SichtbarkeitSetzen(gewicht != "")
        self.text.TextSetzen(gewicht)
        self.DarstellungAktualisieren()

    # Aktualisiert das Symbol
    def DarstellungAktualisieren(self):
        x1 = self.von.XGeben()
        y1 = self.von.YGeben()
        x2 = self.nach.XGeben()
        y2 = self.nach.YGeben()
        laenge = int(math.sqrt((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2)))
        xm = int((x1 + x2) / 2.0)
        ym = int((y1 + y2) / 2.0)
        x = int(xm - laenge / 2)
        y = int(ym - self.breite / 2)
        self.aussen.PositionSetzen(x, y)
        self.aussen.GroesseSetzen(laenge, self.breite)
        if (self.breite < 4):
            groesse = self.breite * 4
            self.innen.SichtbarkeitSetzen(False)
            self.aussen.FarbeSetzen(self.farbe)
            self.pfeil.SichtbarkeitSetzen(self.gerichtet)
            self.pfeil.PositionSetzen(xm, int(ym - groesse / 2))
            self.pfeil.GroesseSetzen(groesse, groesse)
            self.pfeil.WinkelSetzen(270)
        else:
            delta = 2 if self.breite <= 8 else 4
            self.innen.SichtbarkeitSetzen(True)
            self.innen.PositionSetzen(x, int(y + delta / 2))
            self.innen.GroesseSetzen(laenge, self.breite - delta)
            self.innen.FarbeSetzen(self.farbe)
            self.aussen.FarbeSetzen("schwarz")
            self.pfeil.SichtbarkeitSetzen(self.gerichtet)
            self.pfeil.PositionSetzen(xm, int(ym - self.breite / 2))
            self.pfeil.GroesseSetzen(self.breite - delta, self.breite)
            self.pfeil.WinkelSetzen(270)

        winkel = 0.0
        if x1 == x2:
            if y2 > y1:
                winkel = -90
            else:
                winkel = 90
        elif x1 < x2:
            winkel = -math.atan((y2 - y1) / (x2 - x1)) / math.pi * 180.0
        else:
            winkel = 180 - math.atan((y2 - y1) / (x2 - x1)) / math.pi * 180.0
        self.pfeil.WinkelSetzen(int(winkel + 270))
        self.innen.WinkelSetzen(int(winkel))
        self.aussen.WinkelSetzen(int(winkel))
        self.pfeil.GanzNachHintenBringen()
        self.innen.GanzNachHintenBringen()
        self.aussen.GanzNachHintenBringen()
        if ((-90 <= winkel) and (winkel <= 0)):
            w = (winkel + 90) / 180.0 * math.pi
            delta = self.breite
            self.text.PositionSetzen(xm + int(delta * math.cos(w)), ym - int(delta * math.sin(w)))
        elif ((0 < winkel) and (winkel <= 90)):
            w = (winkel + 90) / 180.0 * math.pi
            delta = self.breite
            self.text.PositionSetzen(xm - int(delta * math.cos(w)), ym + int(delta * math.sin(w)) + 7)
        elif ((90 < winkel) and (winkel <= 180)):
            w = (winkel + 90) / 180.0 * math.pi
            delta = self.breite
            self.text.PositionSetzen(xm - int(delta * math.cos(w)), ym + int(delta * math.sin(w)))
        else:
            w = (winkel + 90) / 180.0 * math.pi
            delta = self.breite
            self.text.PositionSetzen(xm + int(delta * math.cos(w)), ym - int(delta * math.sin(w)) + 10)
    
    # Entfernt das Kantensymbol aus der Anzeige
    def Entfernen(self):
        self.aussen.Entfernen()
        self.innen.Entfernen()
        self.pfeil.Entfernen()
        self.text.Entfernen()
    
    # Setzt die Sichtbarkeit des Symbols
    def SichtbarkeitSetzen(self, sichtbar):
        self.aussen.SichtbarkeitSetzen(sichtbar)
        self.innen.SichtbarkeitSetzen(sichtbar and self.breite >= 4)
        self.text.SichtbarkeitSetzen(sichtbar)
        self.pfeil.SichtbarkeitSetzen(sichtbar and self.gerichtet)

    # Setzt die Farbe der Darstellung
    def FarbeSetzen(self, f):
        self.farbe = f
        self.DarstellungAktualisieren()

    # Meldet den Startknoten
    def StartsymbolGeben(self):
        return self.von

    # Meldet den Zielknoten
    def ZielsymbolGeben(self):
        return self.nach


# Beschreibt einen Knoten
class Knoten:

    # Besetzt die Attribute und legt das Knotensymbol an.
    def __init__(self, bezeichner, x, y):
        self.bezeichner = bezeichner
        self.laenge = 0
        self.vorgaenger = -1
        self.symbol = KnotenSymbol(x, y, 20, "weiss", bezeichner)
        self.blaetter = []

    # Meldet den Bezeichner des Knotens zurück.
    def BezeichnerGeben(self):
        return self.bezeichner

    # Meldet das Symbol des Knotens zurück.
    def SymbolGeben(self):
        return self.symbol

    # Setzt die Farbe der Darstellung
    def FarbeSetzen(self, f):
        self.symbol.FarbeSetzen(f)
    
    # Setzt die Weglänge
    def LaengeSetzen(self, l):
        self.laenge = l
    
    # Meldet die Weglänge zurück.
    def LaengeGeben(self):
        return self.laenge
    
    # Meldet den Vorgänger des Knotens zurück.
    def VorgaengerGeben(self):
        return self.vorgaenger
    
    # Setzt den Vorgänger für den Weg zum Ausgang
    def VorgaengerSetzen(self, v):
        self.vorgaenger = v
    
    # Fügt eine neues Blatt für den Breitensuchebaum an
    def NeuesBlattAnfuegen(self, blatt):
        self.blaetter.append(blatt)
    
    # Meldet die Blätter des Beritensuchebaums zurück.
    def BlaetterGeben(self):
        return self.blaetter


# Stellt Methoden zum Lesen der Graphendaten bereit
class Lesen:

    # Liest die Datenbank unter dem gegebenen Namen ein und trägt die Daten in den Graphen g ein.
    def LesenDatenbank(self, name, g):
        try:
            verbindung = sqlite3.connect(name)
            daten = verbindung.execute("SELECT bezeichner, x, y FROM knoten;")
            for datensatz in daten:
                g.KnotenEinfuegen(str(datensatz[0]), int(datensatz[1]), int(datensatz[2]))

            daten = verbindung.execute("SELECT bezeichnerStart, bezeichnerZiel, gewicht, gerichtet FROM kanten;")
            for datensatz in daten:
                g.KanteEinfuegen(str(datensatz[0]), str(datensatz[1]), int(datensatz[2]))

            verbindung.close()
            return True
        except Exception as e:
            print(e)
            return False

    # Liest die Datei unter dem gegebenen Namen ein und trägt die Daten in den Graphen g ein.
    def LesenDatei(self, name, g):
        try:
            with open(name, 'r') as datei:
                zeilen = [zeile.rstrip() for zeile in datei]

                anzahl = int(zeilen[0].split(": ")[1])
                aktuelle_zeile = 1
                for i in range(anzahl):
                    teile = zeilen[aktuelle_zeile].split("\t")
                    g.KnotenEinfuegen(teile[2], int(teile[0]), int(teile[1]))
                    aktuelle_zeile += 1
                anzahl = int(zeilen[aktuelle_zeile].split(": ")[1])
                aktuelle_zeile += 1
                for i in range(anzahl):
                    teile = zeilen[aktuelle_zeile].split("\t")
                    g.KanteEinfuegen(teile[0], teile[1], int(teile[2]))
                    aktuelle_zeile += 1
            return True
        except Exception as e:
            print(e)
            return False
